/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.adl;

import java.util.HashMap;
import java.util.Map;

/**
 * An abstract {@link Node}.
 */

public abstract class AbstractNode implements Node {

  private final String        type;

  private String              source;

  // On-demand allocation of the decoration map to minimize the memory
  // footprint.
  private Map<String, Object> decorations;

  protected AbstractNode(final String type) {
    this.type = type;
  }

  /**
   * Seamless on-demand allocation of the decoration map.
   * 
   * @return allocated decoration map.
   */
  private Map<String, Object> decorations() {
    if (decorations == null) decorations = new HashMap<String, Object>();
    return decorations;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Node interface
  // ---------------------------------------------------------------------------

  public String astGetType() {
    return type;
  }

  public String astGetSource() {
    return source;
  }

  public void astSetSource(final String source) {
    this.source = source;
  }

  public Object astGetDecoration(final String name) {
    if (decorations == null)
      return null;
    else
      return this.decorations().get(name);
  }

  public Map<String, Object> astGetDecorations() {
    if (decorations == null)
      return new HashMap<String, Object>();
    else
      return new HashMap<String, Object>(this.decorations());
  }

  public void astSetDecoration(final String name, final Object decoration) {
    this.decorations().put(name, decoration);
  }

  public void astSetDecorations(final Map<String, Object> decorations) {
    if (decorations.size() > 0) this.decorations().putAll(decorations);
  }

  // ---------------------------------------------------------------------------
  // Overridden methods
  // ---------------------------------------------------------------------------

  @Override
  public String toString() {
    // look for a "name" attribute
    final Map<String, String> attributes = astGetAttributes();
    final String name = attributes.get("name");
    if (name != null) {
      return type + '<' + name + '>';
    } else if (attributes.size() == 1) {
      // node contains only one attribute
      final Map.Entry<String, String> attr = attributes.entrySet().iterator()
          .next();
      return type + '<' + attr.getKey() + '=' + attr.getValue() + '>';
    } else {
      return type + '@' + System.identityHashCode(this);
    }
  }
}
